
{-
Copyright 2009, Jitesh Jain, Shashank Kapoor, Sajal Kumar Sinha, Piyush Srivastava and Utkarsh Upadhyay.

This file is part of atp, an automated theorem prover.

atp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

atp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with atp.  If not, see <http://www.gnu.org/licenses/>.

-}


{-# OPTIONS_GHC -XTypeSynonymInstances -cpp #-}
module Logic1 where
import Data.Char
import Data.List as List
import Data.Map as Map
import Control.Monad
import Text.ParserCombinators.Parsec
import qualified Text.ParserCombinators.Parsec.Token as P


data CNF        =       CNF {cnfName::String, role::String, formula::CnfFormula} deriving (Eq, Show)

type CnfFormula =       [Literal]

data Literal    =       Not AtomicFormula | Pos AtomicFormula 
                        | Eq Term Term | Neq Term Term 
{-
#ifdef DEBUG
-}
                        deriving (Eq, Ord , Show) 
{-
#else
-}	
                        deriving (Eq, Ord)
{-
#endif
-}
data AtomicFormula   =  AtomicFormula {afName::LowerWord, afArg:: [Term]}
{-
#ifdef DEBUG
-}
                        deriving (Eq, Ord , Show) 
{-
#else
-}			
                        deriving (Eq, Ord)
{-
#endif
-}


                                -- a predicate with a list of terms
                  
data Term       =       Function {tName::LowerWord, tArg:: [Term]}  -- Constants are functions of zero arguments
                        | Var UpperWord | VarT UpperWord1 
{-
#ifdef DEBUG
-}
			deriving (Eq, Ord , Show) 
{-
#else
-}
			deriving (Eq, Ord)
{-
#endif
-}


              
--User State for the parser
data LState = LState {lSLowest::Int, lSTable::Map.Map String Int}
initialState :: LState
initialState = LState{lSLowest = 0, lSTable = Map.empty}
getVariables :: Term -> [UpperWord]
getVariables (Var a) = [a]
getVariables (Function _ listArgs) = uniq [] $ sort $ concatMap getVariables listArgs
	where uniq before after = case after of
				    []		-> before
				    [x]		-> uniq (x:before) []
				    (x:y:rest) 	| x==y 	    -> uniq before (y:rest)
				    		| otherwise -> uniq (x:before) (y:rest)

isVariable :: Term -> Bool
isVariable (Var _) = True
isVariable (Function _ _) = False

{-
#ifndef DEBUG
-}
instance Show Term where
       show (Function {tName=name, tArg=arg}) = case arg of
                            []                  -> name
                            (firstArg:restArg) -> name ++ "(" ++ (foldl (\y x -> y ++ "," ++ (show x)) (show firstArg) restArg) ++ ")"
       show (Var x) = "X_" ++ show x

instance Show AtomicFormula where
	show (AtomicFormula name terms) = case terms of
					       [] -> name
					       (firstArg:restArg) -> name ++ "(" ++ (foldl (\y x -> y ++ "," ++ (show x)) (show firstArg) restArg) ++ ")"

instance Show Literal where
	show (Pos a) = show a
	show (Not a) = "~" ++ show a

showCnf :: CnfFormula -> String
showCnf x = "(" ++ concat (List.intersperse " | " (List.map show x)) ++ ")"
{-
#else
-}
showCnf :: CnfFormula -> String
showCnf = show
{-
#endif
-}

type UpperWord = Int
type UpperWord1 = String
type LowerWord = String


languageDef ::P.LanguageDef LState
languageDef = P.LanguageDef{
                        P.commentStart = "/*"
                        , P.commentEnd = "*/"
                        , P.commentLine = "%"
                        , P.nestedComments = True
                        , P.identStart = char '$' <|> lower
                        , P.identLetter= alphaNum <|> char '_'
                        , P.opStart    = oneOf "~!="
                        , P.opLetter   = oneOf "="
                        , P.reservedOpNames = ["!=", "~", "="]
                        , P.reservedNames = []
                        , P.caseSensitive = True
                }

lexer :: P.TokenParser LState
lexer = P.makeTokenParser languageDef


comma = P.comma lexer
identifier = P.identifier lexer
identLetter = P.identLetter languageDef
integer = P.integer lexer
lexeme = P.lexeme lexer
natural = P.natural lexer
parens = P.parens lexer
reservedOp = P.reservedOp lexer
reserved = P.reserved lexer
semi = P.semi lexer
squares = P.squares lexer
symbol = P.symbol lexer
whiteSpace = P.whiteSpace lexer



--From the TPTP grammar
cnfAnnotated :: CharParser LState CNF
cnfAnnotated = lexeme (
                do{              setState initialState
                        ;        string "cnf"
                        ;        x <- parens (
                                do{
                                                 z <- name; comma
                                        ;        x <- formulaRole; comma
                                        ;        y <- cnfFormula
                                        ;        annotations
                                        ;        return CNF{cnfName = z, role = x, formula = y}
                                })
                        ;       char '.'
                        ;       return x
                }) <?> "CNF formula"

name :: CharParser LState String
name =  atomicWord <|> do{x<-natural; return (show x)} <?> "name"

atomicWord :: CharParser LState String
atomicWord = lexeme(
                (do{
                                  x <- lower
                        ;         y <- many identLetter
                        ;         return (x:y)
                } <?> "lower case word") 
                <|>
                (do{
                                char '\''
                        ;        y <- many1 sqChar
                        ;        char '\''
                        ;         return y
                } <?> "single quoted string"))

sqChar        :: CharParser LState Char
sqChar  = satisfy (\s-> let p = ord s in (p >= 32 && p<=38) || (p >= 40 && p<= 126 && p /= 92))
        --Will need to change later, since escaped single quotes are allowed.


formulaRole :: CharParser LState String
formulaRole = lexeme( choice [  string "axiom", string "axiom" ,
                                string "hypothesis" , string "definition" ,
                                string "assumption" , string "lemma" ,
                                string "theorem" , string "conjecture" ,
                                string "negated_conjecture" , string "plain"]) <?> "Formula Role"

annotations :: CharParser LState String
annotations = return [] --This will need to change later

cnfFormula :: CharParser LState CnfFormula
cnfFormula = do{x<-literal; return [x]} <|> (liftM sort $ parens( sepBy1 literal vLine)) <?> "a collection of disjuncts"

vLine :: CharParser LState Char
vLine = lexeme( char '|' ) <?> "vertical bar "

literal :: CharParser LState Literal

literal = lexeme(
                do{
                                char '~'
                        ;       whiteSpace;
                        ;       x <- atomicFormula;
                                return (Not x)
                } <|> (
                try fol_infix_unary <|>
                do{
                                x <- atomicFormula
                        ;       return (Pos x)
                })
                ) <?> "literal"
               
atomicFormula :: CharParser LState AtomicFormula
atomicFormula = lexeme(do{
                                        x <- identifier
                                ;       y <- option [] arguments
                                ;       return AtomicFormula{afName = x, afArg = y}
                              }
                 ) <?> "atomic formula"

arguments :: CharParser LState [Term]
arguments = parens (sepBy1 term comma) 


fol_infix_unary :: CharParser LState Literal
fol_infix_unary = do{
                                x <- term
                        ;       do{
                                                reservedOp "!="
                                        ;       y <- term
                                        ;       return (Neq x y)
                                } 
                                <|>
                                do{
                                                reservedOp "="
                                        ;       y <- term
                                        ;       return (Eq x y)
                                }
                    }


term :: CharParser LState Term
term    =  try functionTerm <|> variable <?> "term"

variable :: CharParser LState Term
variable  = lexeme (do{x <- upper
                       ; y <- many identLetter
                       ; modifyVariable ( x:y)})
                   where modifyVariable :: String -> CharParser LState Term
                         modifyVariable s = do state <- getState
                                               let low = lSLowest state
                                                   table = lSTable state
                                                   query = member s table
                                                   in if query 
                                                         then return (Var (table ! s))
                                                         else do setState LState{lSLowest = low + 1,
                                                                        lSTable = Map.insert s low table}
                                                                 return (Var low)


        

functionTerm :: CharParser LState Term
functionTerm = lexeme (do{x<-identifier; y <- option [] arguments; return (Function {tName = x, tArg = y})})

manyCnf :: CharParser LState [CNF]
manyCnf = many cnfAnnotated

runOnFile :: (Show a) => CharParser LState a -> FilePath -> IO ()
runOnFile p x = do input <- readFile x
                   let result = runParser p (LState 0 Map.empty) x input
                   case result of 
                        Right a ->  print a
                        Left err -> error $ show err

getFromFile :: FilePath -> IO [CnfFormula]
getFromFile x = do input <- readFile x
		   let result = runParser (do whiteSpace; many cnfAnnotated) (LState 0 Map.empty) x input
		   case result of 
			Left err -> error (show err)
			Right a -> return (List.map formula a)



